-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("Aloft")
local SML = AceLibrary("SharedMedia-1.0")
local roster = AceLibrary:HasInstance("Roster-2.1") and AceLibrary("Roster-2.1") or nil

-----------------------------------------------------------------------------

Aloft = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "AceDB-2.0", "FuBarPlugin-2.0")
Aloft.version = "1.0"
Aloft.hasIcon = "Interface\\Icons\\Spell_Frost_Wisp"	-- Placeholder
Aloft.defaultPosition = 'LEFT'
Aloft.hideWithoutStandby = true
Aloft.independentProfile = true

-----------------------------------------------------------------------------

local function FormatLargeValue(value)
	if value < 10000 then
		-- Force it to an integer
		return string.format("%d", value)
	elseif value < 100000 then
		return string.format("%.1fk", value / 1000)
	elseif value < 1000000 then
		return string.format("%dk", value / 1000)
	else
		return string.format("%.2fm", value / 1000000)
	end
end



local numberOfWorldChildren = 0
local nameplateList = { }
local outlineLookup = { ["None"] = nil, ["Normal"] = "OUTLINE", ["Thick"] = "THICKOUTLINE" }
local healthTextMethods =
{
	["Hidden"] = function(frame, healthBar) return "" end,
	["Percent"] = function(frame, healthBar) return string.format("%d%%", floor(100 * healthBar:GetValue() / select(2, healthBar:GetMinMaxValues()))) end,
	["Group Health"] =	function(frame, healthBar) 
							local maxValue = select(2, healthBar:GetMinMaxValues())
							if maxValue == 100 then 
								return "" 
							end
							return FormatLargeValue(healthBar:GetValue())
						end,
	["Group Health Deficit"] =	function(frame, healthBar) 
									local maxValue = select(2, healthBar:GetMinMaxValues())
									if maxValue == 100 then 
										return "" 
									end
									local value = healthBar:GetValue()
									if value == maxValue then 
										return "" 
									end
									return FormatLargeValue(maxValue - value)
								end,
	["Group Health"] =	function(frame, healthBar) 
							local maxValue = select(2, healthBar:GetMinMaxValues())
							if maxValue == 100 then 
								return ""
							end
							return FormatLargeValue(healthBar:GetValue())
						end,
	["Group Health Deficit"] =	function(frame, healthBar) 
									local maxValue = select(2, healthBar:GetMinMaxValues())
									if maxValue == 100 then 
										return "" 
									end
									local value = healthBar:GetValue()
									if value == maxValue then 
										return "" 
									end
									return FormatLargeValue(maxValue - value)
								end,
	["Health"] =	function (frame, healthBar)
						local maxValue = select(2, healthBar:GetMinMaxValues())
						if maxValue ~= 100 then 
							return FormatLargeValue(healthBar:GetValue())
						end

						if MobHealth3 then
							local health, maxHealth, valid = MobHealth3:GetHealth(healthBar:GetValue(), maxValue, frame.aloftData.name, frame.aloftData.level)
							if valid then
								return FormatLargeValue(health)
							end							
						elseif MobHealthDB then
							local mobHealthIndex = string.format("%s:%d", frame.aloftData.name, frame.aloftData.level)
							local mobHealthData = MobHealthDB[mobHealthIndex]
							if mobHealthData then
								local s, e, pts, pct = string.find(mobHealthData, "^(%d+)/(%d+)$")
								if( pts and pct ) then
									pct = tonumber(pct)
									if( pct ~= 0 ) then
										return FormatLargeValue((healthBar:GetValue() * tonumber(pts) / pct))
									end
								end							
							end
						end
						
						return string.format("%d%%", healthBar:GetValue())
					end
}

-----------------------------------------------------------------------------

function Aloft:OnInitialize()
	self:SetText(L["Aloft"])
end

function Aloft:OnEnable()
	self:ScheduleRepeatingEvent("AloftUpdate", self.Update, 0.2, self)
	self:RegisterEvent("SharedMedia_SetGlobal", function(mediatype, override)
		if mediatype == "statusbar" then
			self:UpdateHealthBars()
			self:UpdateHighlights()
			self:UpdateOverlays()
		end
	end)
end

function Aloft:OnDisable()
end

function Aloft:IsNameplateFrame(frame)
	if frame:GetName() then
		return false
	end

	local overlayFrame = select(1, frame:GetRegions())	
	if not overlayFrame or overlayFrame:GetObjectType() ~= "Texture" or overlayFrame:GetTexture() ~= "Interface\\Tooltips\\Nameplate-Border" then
		return false
	end
	
	return true
end

function Aloft:ProcessChildFrames(...)
	for i = 1, select('#', ...) do
		local frame = select(i, ...)
		if self:IsNameplateFrame(frame) then 
			if not nameplateList[frame] then
				self:SetupNamePlate(frame)
				nameplateList[frame] = frame
			end
		end
	end
end

function Aloft:Update()
	local newNumberOfWorldChildren = WorldFrame:GetNumChildren()
	if newNumberOfWorldChildren ~= numberOfWorldChildren then
		numberOfWorldChildren = newNumberOfWorldChildren

		self:ProcessChildFrames(WorldFrame:GetChildren())
	end
end

if Prat_PlayerNames then
	Aloft.PratPlayerNamesUpdateMouseOver = Prat_PlayerNames.updateMouseOver
	Prat_PlayerNames.updateMouseOver = function()
		if UnitIsPlayer("mouseover") and UnitIsFriend("player", "mouseover") and not Prat_PlayerNames.Classes[UnitName("mouseover")] then
			Aloft.PratPlayerNamesUpdateMouseOver(Prat_PlayerNames)
			Aloft:UpdateHealthBarColourForUnitNamed(UnitName("mouseover"))
		end
	end

end

function Aloft:UpdateHealthBarColourForUnitNamed(name)
	for frame in pairs(nameplateList) do	
		if frame.aloftData.name == name then
			self:UpdateNameplate(frame)
		end
	end
end

-----------------------------------------------------------------------------

function Aloft:PlaceFrame(frame, healthBar, placement)
	if placement.visible then
		frame:ClearAllPoints()
		frame:SetPoint(placement.point, healthBar, placement.relativeToPoint, placement.offsetX, placement.offsetY)
		
		if placement.size then
			frame:SetWidth(placement.size)
			frame:SetHeight(placement.size)
		end
		frame:SetAlpha(placement.alpha or 1.0)
	else
		frame:SetAlpha(0)
	end
end

-----------------------------------------------------------------------------

function Aloft:UpdateHealthBarColor(frame, healthBar)
	local overlayFrame, frame2, frame3, highlightFrame, nameFrame, levelTextFrame, bossIconFrame, raidIconFrame = frame:GetRegions()

--	min,max = healthBar:GetMinMaxValues()
--	ChatFrame3:AddMessage("HB: "..frame.aloftData.name.." - "..min.."/"..max)

--	if frame.aloftData.nameFrameText ~= nameFrame:GetText() then
--		ChatFrame3:AddMessage("Data mismatch in UpdateHealthBarColor: "..tostring(frame.aloftData.name).."/"..tostring(nameFrame:GetText()))
--		ChatFrame3:AddMessage(debugstack())
--	end

	if self.db.profile.healthBar.colorByClass then
		if frame.aloftData.type == "friendlyPlayer" then
			if not self.db.profile.healthBar.groupOnly and Prat_PlayerNames and Prat_PlayerNames.Classes[frame.aloftData.name] then		   
				local color = RAID_CLASS_COLORS[string.upper(Prat_PlayerNames.Classes[frame.aloftData.name])]
				if color then
					healthBar:SetStatusBarColor(color.r, color.g, color.b, self.db.profile.healthBar.alpha)
					return
				end
			end
			if roster then
				local unitid = roster:GetUnitIDFromName(frame.aloftData.name)
				if unitid then
					local color = RAID_CLASS_COLORS[select(2, UnitClass(unitid))]
					if color then
						healthBar:SetStatusBarColor(color.r, color.g, color.b, self.db.profile.healthBar.alpha)
						return
					end
				end
			end
		elseif frame.aloftData.type == "friendlyNPC" then
			if roster and roster:GetUnitIDFromName(frame.aloftData.name) then
				-- Must be a party pet
				local color = self.db.profile.healthBar.colors.partyPet
				healthBar:SetStatusBarColor(color[1], color[2], color[3], self.db.profile.healthBar.alpha)
				return
			end
		end
	end
	
	local color = self.db.profile.healthBar.colors[frame.aloftData.type]
	if color then
		healthBar:SetStatusBarColor(color[1], color[2], color[3], self.db.profile.healthBar.alpha)
	else
		local r, g, b = healthBar:GetStatusBarColor()
		healthBar:SetStatusBarColor(r, g, self.db.profile.healthBar.alpha)
	end
end

function Aloft:UpdateHealthText(frame, healthBar)
	local healthTextFrame = frame.aloftData.healthTextFrame
	
	if healthTextFrame then
		healthTextFrame:SetText(healthTextMethods[self.db.profile.healthText.mode](frame, healthBar))
	end
end

function Aloft:UpdateNameplate(frame)
	local healthBar = frame:GetChildren()

	self:UpdateHealthBarColor(frame, healthBar)
	self:UpdateHealthText(frame, healthBar)
end

function Aloft:OnHealthBarValueChanged(frame)
	local _, _, _, _, nameFrame = frame:GetRegions()
	
	if frame.aloftData.nameFrameText ~= nameFrame:GetText() then
		self:OnNameplateShow(frame)
	else
		self:UpdateNameplate(frame)
	end
end

function Aloft:OnNameplateShow(frame)
	-- When a nameplate is first shown, grab some data about it
	
	local overlayFrame, frame2, frame3, highlightFrame, nameFrame, levelTextFrame, bossIconFrame, raidIconFrame = frame:GetRegions()
	local healthBar = frame:GetChildren()

	local aloftData = frame.aloftData

	if aloftData.nameFrameText ~= nameFrame:GetText() then
		aloftData.name = nameFrame:GetText()
	end
	aloftData.nameFrameText = aloftData.name
	aloftData.isBoss = bossIconFrame:IsShown()
	aloftData.level = tonumber(levelTextFrame:GetText()) or -1
	if not levelTextFrame:IsShown() then
		aloftData.level = -1
	end

	local r, g, b = healthBar:GetStatusBarColor()
	r, g, b = floor(r * 256), floor(g*256), floor(b*256)

	if r == 255 and g == 0 and b == 0 then
		if aloftData.isBoss then
			aloftData.type = "hostileBoss"
		else
			aloftData.type = "hostile"
		end
	elseif r == 0 and g == 0 and b == 255 then
		aloftData.type = "friendlyPlayer"
	elseif r == 0 and g == 255 and b == 0 then
		if aloftData.isBoss then
			aloftData.type = "friendlyBoss"
		else
			aloftData.type = "friendlyNPC"
		end
	elseif r == 255 and g == 255 and b == 0 then
		aloftData.type = "neutral"
	else
--		ChatFrame3:AddMessage("rgb = "..r.."."..g.."."..b)
	end

	if self.db.profile.nameplateText.overrideColors then
		local color = self.db.profile.nameplateText.colors[aloftData.type]
	
		if color then
			aloftData.nameFrameText = string.format("|c%02x%02x%02x%02x%s", color[4]*255, color[1]*255, color[2]*255, color[3]*255, aloftData.name)
			nameFrame:SetText(aloftData.nameFrameText)
		end
	
	end
--	ChatFrame3:AddMessage(aloftData.name..": "..aloftData.level.. (aloftData.isBoss  and " boss" or " not boss"))

	if not self.db.profile.levelText.visible then
		levelTextFrame:Hide()
	end

	self:PlaceFrame(bossIconFrame, healthBar, self.db.profile.bossIcon)
	highlightFrame:SetAllPoints(healthBar)

	-- The frame has an active region for detecting hits that is given roughly by this formula.
	frame:SetHeight(healthBar:GetHeight()*1.6)
	frame:SetWidth(healthBar:GetWidth()*1.05)
	
	self:UpdateNameplate(frame)
end

function Aloft:OnNameplateHide(frame)
	frame.aloftData.name = "[invalid]"
	frame.aloftData.type = "unknown"
end

-----------------------------------------------------------------------------

local function OnNameplateShow()
	Aloft:OnNameplateShow(this)
end

local function OnNameplateHide()
	Aloft:OnNameplateHide(this)
end

local function OnHealthBarValueChanged()
	Aloft:OnHealthBarValueChanged(this:GetParent())
end

-----------------------------------------------------------------------------

function Aloft:SetupHealthTextFrame(frame, healthBar)
	if self.db.profile.healthText.mode == "Hidden" then
		return
	end
	
	local healthTextFrame = frame.aloftData.healthTextFrame
	
	if not healthTextFrame then
		healthTextFrame = frame:CreateFontString()
		frame.aloftData.healthTextFrame = healthTextFrame
	end
	
	local _, _, _, _, nameFrame = frame:GetRegions()
	healthTextFrame:SetFont(nameFrame:GetFont(), self.db.profile.healthText.fontSize, outlineLookup[self.db.profile.healthText.outline])
	healthTextFrame:SetTextColor(unpack(self.db.profile.healthText.color))
	self:PlaceFrame(healthTextFrame, healthBar, self.db.profile.healthText)
end

function Aloft:SetupNamePlate(frame)
	frame.aloftData = { type = "unknown" }

	local overlayFrame, frame2, frame3, highlightFrame, nameFrame, levelTextFrame, bossIconFrame, raidIconFrame = frame:GetRegions()
	local healthBar = frame:GetChildren()

	highlightFrame:SetAllPoints(healthBar)
	highlightFrame:SetVertexColor(unpack(self.db.profile.highlight.color))	
	highlightFrame:SetTexture(SML:Fetch("statusbar", self.db.profile.highlight.texture))
	highlightFrame:SetBlendMode(self.db.profile.highlight.blendMode)
	
	nameFrame:ClearAllPoints()
	nameFrame:SetPoint("TOPLEFT", healthBar, "TOPLEFT", self.db.profile.nameplateText.offsets.left, self.db.profile.nameplateText.offsets.vertical)
	nameFrame:SetPoint("BOTTOMRIGHT", healthBar, "BOTTOMRIGHT", self.db.profile.nameplateText.offsets.right, self.db.profile.nameplateText.offsets.vertical)
	nameFrame:SetFont(nameFrame:GetFont(), self.db.profile.nameplateText.fontSize, outlineLookup[self.db.profile.nameplateText.outline])
	nameFrame:SetJustifyH(self.db.profile.nameplateText.alignment)
	nameFrame:SetShadowOffset(0, 0)

	self:PlaceFrame(levelTextFrame, healthBar, self.db.profile.levelText)
	levelTextFrame:SetFont(levelTextFrame:GetFont(), self.db.profile.levelText.fontSize, outlineLookup[self.db.profile.levelText.outline])
	levelTextFrame:SetShadowOffset(0, 0)

	self:PlaceFrame(raidIconFrame, healthBar, self.db.profile.raidIcon)
	self:PlaceFrame(bossIconFrame, healthBar, self.db.profile.bossIcon)

	overlayFrame:SetAllPoints(healthBar)
	overlayFrame:SetTexture(SML:Fetch("statusbar", self.db.profile.overlay.texture))
	overlayFrame:SetVertexColor(unpack(self.db.profile.overlay.color))
	
	if not self.db.profile.overlay.visible then	
		overlayFrame:Hide()
	end
	
	if not self.db.profile.nameplateText.visible then
		nameFrame:Hide()
	end

	self:SetupHealthTextFrame(frame, healthBar)
	
	frame:SetScript("OnShow", OnNameplateShow)
	frame:SetScript("OnHide", OnNameplateHide)

	healthBar:SetScript("OnValueChanged", OnHealthBarValueChanged)
	healthBar:SetStatusBarTexture(SML:Fetch("statusbar", self.db.profile.healthBar.texture))

	self:OnNameplateShow(frame)
end

-----------------------------------------------------------------------------
-- Options methods. All of these need access to nameplateList
-----------------------------------------------------------------------------

function Aloft:UpdateNameplateText()
	for frame in pairs(nameplateList) do	
		local _, _, _, _, nameFrame = frame:GetRegions()
		local healthBar = frame:GetChildren()
	
		if self.db.profile.nameplateText.visible then
			nameFrame:Show()
		else
			nameFrame:Hide()
		end
	
		nameFrame:ClearAllPoints()
		nameFrame:SetPoint("TOPLEFT", healthBar, "TOPLEFT", self.db.profile.nameplateText.offsets.left, self.db.profile.nameplateText.offsets.vertical)
		nameFrame:SetPoint("BOTTOMRIGHT", healthBar, "BOTTOMRIGHT", self.db.profile.nameplateText.offsets.right, self.db.profile.nameplateText.offsets.vertical)
		nameFrame:SetFont(nameFrame:GetFont(), self.db.profile.nameplateText.fontSize, outlineLookup[self.db.profile.nameplateText.outline])
		nameFrame:SetJustifyH(self.db.profile.nameplateText.alignment)
		
		if frame:IsShown() then
			self:OnNameplateShow(frame)
		end
	end
end

function Aloft:UpdateHealthBars()
	for frame in pairs(nameplateList) do	
		local overlayFrame = frame:GetRegions()
		local healthBar = frame:GetChildren()
		healthBar:SetStatusBarTexture(SML:Fetch("statusbar", self.db.profile.healthBar.texture))
		if frame.aloftData.name ~= "[Invalid]" then
			self:UpdateHealthBarColor(frame, healthBar)
		end
	end
end

function Aloft:UpdateHealthTexts()
	for frame in pairs(nameplateList) do
		local healthBar = frame:GetChildren()
		self:SetupHealthTextFrame(frame, healthBar)
		self:UpdateHealthText(frame, healthBar)
	end
end

function Aloft:UpdateLevelText()
	for frame in pairs(nameplateList) do
		local _, _, _, _, _, levelTextFrame = frame:GetRegions()
		local healthBar = frame:GetChildren()
		
		if self.db.profile.levelText.visible then
			levelTextFrame:Show()
		else
			levelTextFrame:Hide()
		end
		levelTextFrame:SetFont(levelTextFrame:GetFont(), self.db.profile.levelText.fontSize, outlineLookup[self.db.profile.levelText.outline])
		self:PlaceFrame(levelTextFrame, healthBar, self.db.profile.levelText)
	end
end

function Aloft:UpdateRaidIcons()
	for frame in pairs(nameplateList) do	
		local _, _, _, _, _, _, _, raidIconFrame = frame:GetRegions()
		local healthBar = frame:GetChildren()
	
		Aloft:PlaceFrame(raidIconFrame, healthBar, self.db.profile.raidIcon)
	end
end

function Aloft:UpdateBossIcons()
	for frame in pairs(nameplateList) do	
		local _, _, _, _, _, _, bossIconFrame = frame:GetRegions()
		local healthBar = frame:GetChildren()

		Aloft:PlaceFrame(bossIconFrame, healthBar, self.db.profile.bossIcon)
	end
end

function Aloft:UpdateOverlays()
	for frame in pairs(nameplateList) do
		local overlayFrame = frame:GetRegions()
		if self.db.profile.overlay.visible then
			overlayFrame:Show()
		else
			overlayFrame:Hide()
		end
		overlayFrame:SetVertexColor(unpack(self.db.profile.overlay.color))
		overlayFrame:SetTexture(SML:Fetch("statusbar", self.db.profile.overlay.texture))
	end
end

function Aloft:UpdateHighlights()
	for frame in pairs(nameplateList) do
		local _, _, _, highlightFrame = frame:GetRegions()
		highlightFrame:SetVertexColor(unpack(self.db.profile.highlight.color))
		highlightFrame:SetTexture(SML:Fetch("statusbar", self.db.profile.highlight.texture))
		highlightFrame:SetBlendMode(self.db.profile.highlight.blendMode)
	end
end

-----------------------------------------------------------------------------
